module.exports = {
    "extends": "airbnb-base",
    "env": {
        "browser": true,
        "es6": true,
        "jquery": true
    },
    "globals": {
        "$":false,
        "async":false,
        "document":false,
        "fabric": false,
        "Noty": false
    }   
};