#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['a2d_diary']

package_data = \
{'': ['*'],
 'a2d_diary': ['input/*',
               'input\\1_diaries_to_create/*',
               'input\\1_diaries_to_create\\resources/*',
               'input\\2_template_to_encode/*',
               'input\\3_diaries_to_encode/*',
               'output/*',
               'output\\created_diaries/*',
               'output\\created_diaries\\a5_references/*',
               'output\\encoded_diaries/*',
               'output\\temporal/*',
               'static/*',
               'static\\downloads/*',
               'static\\fonts/*']}

install_requires = \
['opencv_python>=3.3.0.10',
 'waitress>=1.0.2',
 'falcon>=1.3.0',
 'numpy>=1.12.1',
 'falcon_multipart>=0.2.0',
 'Pillow>=4.3.0',
 'PyPDF2>=1.26.0',
 'reportlab>=3.4.0']

entry_points = \
{'console_scripts': ['a2d_diary = a2d_diary.__main__:main']}

setup(name='a2d_diary',
      version='0.1.4',
      description='A2D-Diary web app. Create and encode paper diaries automatically',
      author='Julio Vega',
      author_email='julio.vega@protonmail.com',
      url='https://a2d-diary.netlify.com',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3',
     )
